import { world, system, BlockPermutation, EquipmentSlot, ItemStack } from "@minecraft/server"

world.beforeEvents.worldInitialize.subscribe(initEvent => {
    initEvent.blockComponentRegistry.registerCustomComponent("zpig:nether_reactor_core_component", {
        onPlace: (event) => {
            const { block, dimension } = event;
            let invasionEntityExistsAtBlockLocation = false;
            const entitiesAtBlockLocation = dimension.getEntitiesAtBlockLocation(block.location);
            entitiesAtBlockLocation.forEach((entity) => {
                if (entity.typeId == "zpig:invasion_entity") {
                    invasionEntityExistsAtBlockLocation = true;
                }
            });
            if (!invasionEntityExistsAtBlockLocation) {
                dimension.spawnEntity("zpig:invasion_entity", block.location);
            }
        },
        onTick: (event) => {
            const { block, dimension } = event;
            let invasionEntityExistsAtBlockLocation = false;
            const entitiesAtBlockLocation = dimension.getEntitiesAtBlockLocation(block.location);
            entitiesAtBlockLocation.forEach((entity) => {
                if (entity.typeId == "zpig:invasion_entity") {
                    invasionEntityExistsAtBlockLocation = true;
                }
            });
            if (!invasionEntityExistsAtBlockLocation) {
                dimension.spawnEntity("zpig:invasion_entity", block.location);
            }
        }
    });
    initEvent.blockComponentRegistry.registerCustomComponent("zpig:explosive_fungus_crop_growth_logic", {
        onRandomTick({ block }) {
            const growthChance = 1 / 5;
            if (Math.random() > growthChance) return;
            const growth = block.permutation.getState("zpig:explosive_fungus_crop_growth");
            block.setPermutation(block.permutation.withState("zpig:explosive_fungus_crop_growth", growth + 1));
        },
        onPlayerInteract({ block, dimension, player }) {
            if (!player) return;
            const equippable = player.getComponent("minecraft:equippable");
            if (!equippable) return;
            const mainhand = equippable.getEquipmentSlot(EquipmentSlot.Mainhand);
            if (!mainhand.hasItem() || mainhand.typeId !== "minecraft:bone_meal") return;
            if (player.getGameMode() == "creative") {
                block.setPermutation(block.permutation.withState("zpig:explosive_fungus_crop_growth", 3));
            }
            else {
                let growth = block.permutation.getState("zpig:explosive_fungus_crop_growth");
                growth += Math.floor(Math.random() * 2);
                block.setPermutation(block.permutation.withState("zpig:explosive_fungus_crop_growth", growth));
                if (mainhand.amount > 1) mainhand.amount--;
                else mainhand.setItem(undefined);
            }
            const effectLocation = block.center();
            dimension.playSound("item.bone_meal.use", effectLocation);
            dimension.spawnParticle("minecraft:crop_growth_emitter", effectLocation);
        },
    });
    initEvent.blockComponentRegistry.registerCustomComponent("zpig:reinforced_basalt_lock_component", {
        onPlayerInteract: (event) => {
            const { block, dimension, player } = event;
            const playerItemSlot = player.getComponent("minecraft:equippable").getEquipmentSlot(EquipmentSlot.Mainhand);
            const playerItem = playerItemSlot.getItem();
            if (playerItem != undefined && block.permutation.getState("zpig:lock_type") == 0 && block.permutation.getState("zpig:lock_state") == 1 && playerItem.typeId == "zpig:barbarian_key") {
                const queue = [];
                queue.push(block);
                while (queue.length > 0) {
                    const currentBlock = queue.shift();
                    if (currentBlock.typeId == "zpig:reinforced_basalt" || currentBlock.typeId == "zpig:reinforced_basalt_lock") {
                        if (dimension.getBlock({ x: currentBlock.location.x + 1, y: currentBlock.location.y, z: currentBlock.location.z }).typeId == "zpig:reinforced_basalt") {
                            queue.push(dimension.getBlock({ x: currentBlock.location.x + 1, y: currentBlock.location.y, z: currentBlock.location.z }));
                        }
                        if (dimension.getBlock({ x: currentBlock.location.x - 1, y: currentBlock.location.y, z: currentBlock.location.z }).typeId == "zpig:reinforced_basalt") {
                            queue.push(dimension.getBlock({ x: currentBlock.location.x - 1, y: currentBlock.location.y, z: currentBlock.location.z }));
                        }
                        if (dimension.getBlock({ x: currentBlock.location.x, y: currentBlock.location.y, z: currentBlock.location.z + 1 }).typeId == "zpig:reinforced_basalt") {
                            queue.push(dimension.getBlock({ x: currentBlock.location.x, y: currentBlock.location.y, z: currentBlock.location.z + 1 }));
                        }
                        if (dimension.getBlock({ x: currentBlock.location.x, y: currentBlock.location.y, z: currentBlock.location.z - 1 }).typeId == "zpig:reinforced_basalt") {
                            queue.push(dimension.getBlock({ x: currentBlock.location.x, y: currentBlock.location.y, z: currentBlock.location.z - 1 }));
                        }
                        if (dimension.getBlock({ x: currentBlock.location.x, y: currentBlock.location.y + 1, z: currentBlock.location.z  }).typeId == "zpig:reinforced_basalt") {
                            queue.push(dimension.getBlock({ x: currentBlock.location.x, y: currentBlock.location.y + 1, z: currentBlock.location.z }));
                        }
                        if (dimension.getBlock({ x: currentBlock.location.x, y: currentBlock.location.y - 1, z: currentBlock.location.z }).typeId == "zpig:reinforced_basalt") {
                            queue.push(dimension.getBlock({ x: currentBlock.location.x, y: currentBlock.location.y - 1, z: currentBlock.location.z }));
                        }
                        dimension.spawnParticle("zpig:reinforced_basalt_break", currentBlock.center());
                        if (currentBlock.typeId == "zpig:reinforced_basalt_lock") {
                            dimension.playSound("vault.insert_item", currentBlock.center());
                        }
                        else {
                            dimension.playSound("dig.basalt", currentBlock.center());
                        }
                        if (player.getGameMode() == "survival" || player.getGameMode() == "adventure") {
                            if (playerItem.amount > 1) {
                                playerItemSlot.setItem(new ItemStack("zpig:barbarian_key", playerItem.amount - 1));
                            }
                            else {
                                playerItemSlot.setItem(new ItemStack("minecraft:air")); 
                            }
                        }
                        currentBlock.setPermutation(BlockPermutation.resolve("minecraft:air"));
                    }
                }
            }
        },
        onTick: (event) => {
            const { block, dimension } = event;
            const playersList = dimension.getPlayers({location: block.location, maxDistance: 4, excludeGameModes: ['spectator'],});
            if (playersList.length > 0 && block.permutation.getState("zpig:lock_state") == 0) {
                block.setPermutation(block.permutation.withState("zpig:lock_state", 1));
                dimension.playSound("vault.activate", block.center());
                playersList.forEach((player) => {
                    if (block.permutation.getState("zpig:lock_type") == 0) {
                        player.sendMessage("You need a §6Barbarian Key§r to unlock this door.");
                    }
                });
            }
            else if (playersList.length <= 0 && block.permutation.getState("zpig:lock_state") == 1) {
                block.setPermutation(block.permutation.withState("zpig:lock_state", 0));
                dimension.playSound("vault.deactivate", block.center());
            }
        }
    });
    initEvent.blockComponentRegistry.registerCustomComponent("zpig:piglin_conquest_structure_spawner_block", {
        onPlayerInteract: (event) => {
            const { block, dimension, player } = event;
            if (block.permutation.getState("zpig:structure") < 1) {
                block.setPermutation(block.permutation.withState("zpig:structure", block.permutation.getState("zpig:structure") + 1));
            }
            else {
                block.setPermutation(block.permutation.withState("zpig:structure", 0));
            }
            const playersList = dimension.getPlayers({location: block.location, maxDistance: 2, excludeGameModes: ['spectator'],});
            if (playersList.length > 0) {
                playersList.forEach((player) => {
                    player.sendMessage("Structure Number: " + block.permutation.getState("zpig:structure"));
                });
            }
        },
        onTick: (event) => { //comment this code out when saving the structure of the spawner block
            const { block, dimension } = event;
            if (block.permutation.getState("zpig:structure") == 1 && block.isValid()) {
                block.setPermutation(block.permutation.withState("zpig:structure", 0));
                world.structureManager.place("barbarian_arena", dimension, { x: block.location.x - 28, y: block.location.y, z: block.location.z - 24});
            }
        }
    });
    initEvent.blockComponentRegistry.registerCustomComponent("zpig:piglin_conquest_mob_spawner_block", {
        onPlayerInteract: (event) => {
            const { block, dimension, player } = event;
            if (block.permutation.getState("zpig:mob") < 2) {
                block.setPermutation(block.permutation.withState("zpig:mob", block.permutation.getState("zpig:mob") + 1));
            }
            else {
                block.setPermutation(block.permutation.withState("zpig:mob", 0));
            }
            const playersList = dimension.getPlayers({location: block.location, maxDistance: 2, excludeGameModes: ['spectator'],});
            if (playersList.length > 0) {
                playersList.forEach((player) => {
                    player.sendMessage("Mob Number: " + block.permutation.getState("zpig:mob"));
                });
            }
        },
        onTick: (event) => {
            const { block, dimension } = event;
            mobSpawnerBlockLogic(block, dimension);
        },
        onPlace: (event) => {
            const { block, dimension } = event;
            mobSpawnerBlockLogic(block, dimension);
        }
    });
});

function mobSpawnerBlockLogic(block, dimension) { //comment this code out when saving the structure of the spawner block
    if (block.permutation.getState("zpig:mob") == 1 && block.isValid()) {
        block.setPermutation(block.permutation.withState("zpig:mob", 0));
        block.setPermutation(BlockPermutation.resolve("minecraft:air"));
        dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_as_patrol_captain>", block.location);
    }
    else if (block.permutation.getState("zpig:mob") == 2 && block.isValid()) {
        block.setPermutation(block.permutation.withState("zpig:mob", 0));
        block.setPermutation(BlockPermutation.resolve("minecraft:air"));
        dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_in_patrol>", block.location);
    }
}

world.afterEvents.itemReleaseUse.subscribe((data) => {
    const { source, itemStack } = data;
    if (itemStack.typeId == "zpig:dragon_bone_axe") {
        const player = source;
        const axeItem = itemStack;
        const axeList = player.dimension.getEntities({
            type: 'zpig:dragon_bone_axe_projectile',
            location: player.location,
            closest: 1,
        });
        const axe = axeList[0];
        if (axe != undefined) {
            axeItem.getComponent('durability').damage += 1;
            axe.getComponent('inventory').container.setItem(0, axeItem);
            if (player.getGameMode() == "survival" || player.getGameMode() == "adventure") {
                axe.runCommand("event entity @s zpig:survival");
            }
            else {
                axe.runCommand("event entity @s zpig:creative");
            }
            axe.runCommand("event entity @s zpig:player");
        }
    }
});

system.afterEvents.scriptEventReceive.subscribe((data) => {
    if (data.id == "zpig:fill_60th_of_invasion_bar") {
        const invasion = data.sourceEntity;
        if (invasion != undefined) {
            const invasionHealth = invasion.getComponent("health");
            if (invasion.getEffect("health_boost") != undefined) {
                const health_boost_strength = invasion.getEffect("health_boost").amplifier;
                const invasionMaxHealth = (4 * health_boost_strength + 1);
                if ((invasionHealth.currentValue + (invasionMaxHealth / 60)) < invasionMaxHealth) {
                    invasionHealth.setCurrentValue(invasionHealth.currentValue + (invasionMaxHealth / 60));
                }
                else {
                    invasionHealth.resetToMaxValue();
                }
            }
            else {
                invasionHealth.resetToMaxValue();
            }
        }
    }
    else if (data.id == "zpig:dragon_bone_axe_survival_logic") {
        const axe = data.sourceEntity;
        const playerList = axe.dimension.getPlayers({
            location: axe.location,
            closest: 1,
        });
        const player = playerList[0];
        const pInven = player.getComponent('inventory').container;
        if (pInven.emptySlotsCount > 0) {
            const axeItem = axe.getComponent('inventory').container.getItem(0);
            pInven.addItem(axeItem);
            axe.runCommand("playsound random.pop @p ~ ~ ~ 0.5 1.5");
            axe.runCommand("event entity @s zpig:despawn");
        }
    }
});

world.afterEvents.entityDie.subscribe((data) => {
    if (data.deadEntity != undefined && data.deadEntity.typeId == "zpig:invasion_entity") {
        if (data.deadEntity.dimension.getBlock(data.deadEntity.location).typeId == "zpig:finished_nether_reactor_core") {
            data.deadEntity.runCommand("summon zpig:invasion_entity");
        }
    }
});

function ValidLocationForWave(invasion) {
    const spawnRadius = 4;
    let randomX, randomY, randomZ;
    let count = 0;
    do {
        randomX = Math.floor(invasion.location.x + (Math.random() * 2 - 1) * spawnRadius);
        randomY = Math.floor(invasion.location.y + 2 + (Math.random() * 2 - 1));
        randomZ = Math.floor(invasion.location.z + (Math.random() * 2 - 1) * spawnRadius);
        count++;
    } while (!((invasion.dimension.getBlock({ x: randomX, y: randomY, z: randomZ }).isAir &&
    invasion.dimension.getBlock({ x: randomX, y: randomY + 1, z: randomZ }).isAir &&
    invasion.dimension.getBlock({ x: randomX, y: randomY + 2, z: randomZ }).isAir &&
    invasion.dimension.getBlock({ x: randomX + 1, y: randomY, z: randomZ }).isAir &&
    invasion.dimension.getBlock({ x: randomX + 1, y: randomY + 1, z: randomZ }).isAir &&
    invasion.dimension.getBlock({ x: randomX + 1, y: randomY + 2, z: randomZ }).isAir &&
    invasion.dimension.getBlock({ x: randomX, y: randomY, z: randomZ + 1 }).isAir &&
    invasion.dimension.getBlock({ x: randomX, y: randomY + 1, z: randomZ + 1 }).isAir &&
    invasion.dimension.getBlock({ x: randomX, y: randomY + 2, z: randomZ + 1 }).isAir &&
    invasion.dimension.getBlock({ x: randomX + 1, y: randomY, z: randomZ + 1 }).isAir &&
    invasion.dimension.getBlock({ x: randomX + 1, y: randomY + 1, z: randomZ + 1 }).isAir &&
    invasion.dimension.getBlock({ x: randomX + 1, y: randomY + 2, z: randomZ + 1 }).isAir &&
    !invasion.dimension.getBlock({ x: randomX, y: randomY - 1, z: randomZ }).isAir &&
    !invasion.dimension.getBlock({ x: randomX + 1, y: randomY - 1, z: randomZ }).isAir &&
    !invasion.dimension.getBlock({ x: randomX, y: randomY - 1, z: randomZ + 1 }).isAir &&
    !invasion.dimension.getBlock({ x: randomX + 1, y: randomY - 1, z: randomZ + 1 }).isAir)
    ) && count <= 30);
    return { x: randomX + 0.5, y: randomY, z: randomZ + 0.5 };
}

function InvasionLogic() {
    const allInvasionsOverworld = world.getDimension("overworld").getEntities({type: 'zpig:invasion_entity',});
    const allInvasionsNether = world.getDimension("nether").getEntities({type: 'zpig:invasion_entity',});
    const allInvasionsEnd = world.getDimension("the_end").getEntities({type: 'zpig:invasion_entity',});
    const allInvasions = allInvasionsOverworld.concat(allInvasionsNether, allInvasionsEnd);
    allInvasions.forEach((invasion) => {
        if (invasion.dimension.getBlock(invasion.location) != undefined && (invasion.dimension.getBlock(invasion.location).typeId != "zpig:nether_reactor_core" && invasion.dimension.getBlock(invasion.location).typeId != "zpig:finished_nether_reactor_core")) {
            const invaders = invasion.dimension.getEntities({tags: ["zpig:in_invasion", "zpig:invasion_id." + invasion.id],});
            invaders.forEach((invader) => {
                invader.removeTag("zpig:in_invasion");
                invader.removeTag("zpig:invasion_id." + invasion.id);
                invader.runCommand("event entity @s zpig:invasion_end");
            });
            invasion.runCommand("event entity @s zpig:despawn");
        }
        if (invasion.getProperty("zpig:invasion_activate") == true) {
            let hordeType = invasion.getProperty("zpig:invasion_horde_type");
            let wave = invasion.getProperty("zpig:invasion_wave");
            let wave_spawned = invasion.getProperty("zpig:wave_spawned");
            let boss_bar_charged = invasion.getProperty("zpig:boss_bar_charge_finished");
            let max_wave_amount = invasion.getProperty("zpig:max_wave_amount");
            let on_final_wave = invasion.getProperty("zpig:on_final_wave");
            if (invasion.dimension.getBlock({ x: invasion.location.x, y: invasion.location.y + 1, z: invasion.location.z }).typeId == "minecraft:obsidian" && invasion.dimension.getBlock({ x: invasion.location.x, y: invasion.location.y + 2, z: invasion.location.z }).typeId != "minecraft:portal" && invasion.dimension.getBlock({ x: invasion.location.x, y: invasion.location.y + 2, z: invasion.location.z }).typeId != "minecraft:fire") {
                invasion.runCommand("setblock ~ ~2 ~ fire keep");
            }
            if (boss_bar_charged == 0) {
                invasion.runCommand("event entity @s zpig:charge_boss_bar");
            }
            else if (wave_spawned == true && boss_bar_charged == 2) {
                const invaders = invasion.dimension.getEntities({tags: ["zpig:in_invasion", "zpig:invasion_id." + invasion.id],});
                let invadersCurrentHealth = 0;
                let invadersMaxHealth = 0;
                invaders.forEach((invader) => {
                    invadersCurrentHealth += invader.getComponent("health").currentValue;
                    invadersMaxHealth += invader.getComponent("health").defaultValue;
                });
                if (invaders.length > 2 && invasion.getProperty("zpig:mobs_remaining_check") != 3) {
                    invasion.setProperty("zpig:mobs_remaining_check", 3);
                    invasion.runCommand("event entity @s zpig:normal_boss_bar");
                }
                else if (invaders.length == 2 && invasion.getProperty("zpig:mobs_remaining_check") != 2) {
                    invasion.setProperty("zpig:mobs_remaining_check", 2);
                    invasion.runCommand("event entity @s zpig:boss_bar_two_remain");
                }
                else if (invaders.length == 1 && invasion.getProperty("zpig:mobs_remaining_check") != 1) {
                    invasion.setProperty("zpig:mobs_remaining_check", 1);
                    invasion.runCommand("event entity @s zpig:boss_bar_one_remains");
                }
                else if (invaders.length == 0 && invasion.getProperty("zpig:mobs_remaining_check") != 0 && !on_final_wave) { //wave defeated
                    invasion.setProperty("zpig:mobs_remaining_check", 0);
                    invasion.runCommand("event entity @s zpig:boss_bar_none_remain");
                }
                else if (invaders.length == 0 && invasion.getProperty("zpig:mobs_remaining_check") != 0 && on_final_wave) { //invasion defeated
                    if (invasion.dimension.getBlock(invasion.location).typeId == "zpig:nether_reactor_core") {
                        invasion.dimension.getBlock(invasion.location).setPermutation(BlockPermutation.resolve("zpig:finished_nether_reactor_core"))
                    }
                    if (invasion.dimension.getBlock({ x: invasion.location.x, y: invasion.location.y + 2, z: invasion.location.z }).typeId == "minecraft:portal") {
                        invasion.runCommand("fill ~ ~2 ~ ~ ~2 ~ air replace portal");
                    }
                    invasion.runCommand("event entity @s zpig:invasion_victory");
                }
                if (invasion.getProperty("zpig:ready_to_spawn_next_wave_after_cooldown")) { //ready to spawn next wave
                    invasion.getComponent("health").setCurrentValue(1);
                    invasion.setProperty("zpig:invasion_wave", wave + 1);
                    invasion.setProperty("zpig:wave_spawned", false);
                    invasion.setProperty("zpig:num_invaders_in_wave", 0);
                    invasion.setProperty("zpig:boss_bar_charge_finished", 0);
                    invasion.setProperty("zpig:ready_to_spawn_next_wave_after_cooldown", false);
                    invasion.runCommand("event entity @s zpig:normal_boss_bar");
                }
                else if (invadersCurrentHealth == 0 && (wave >= max_wave_amount) && !on_final_wave) { //spawn final wave
                    invasion.setProperty("zpig:on_final_wave", true);
                }
                else if ((invasion.getComponent("health").currentValue != invadersCurrentHealth) && (invadersCurrentHealth < invadersMaxHealth || (invasion.getProperty("zpig:num_invaders_in_wave") != invaders.length))) {
                    if (invadersCurrentHealth == 0) {
                        invasion.getComponent("health").setCurrentValue(1);
                    }
                    else {
                        invasion.getComponent("health").setCurrentValue(invadersCurrentHealth);
                    }
                }
            }
            else if (wave == 0 && wave_spawned == false && boss_bar_charged == 2 && !on_final_wave) { //spawn wave 1
                const pig1 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig2 = invasion.dimension.spawnEntity("zpig:piglin_runt<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig3 = invasion.dimension.spawnEntity("zpig:piglin_runt<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig4 = invasion.dimension.spawnEntity("zpig:piglin_runt<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                invasion.dimension.playSound("portal.travel", pig1.location);
                invasion.setProperty("zpig:wave_spawned", true);
                const invaders = [pig1, pig2, pig3, pig4];
                let invadersHealth = 0;
                invaders.forEach((invader) => {
                    invader.addTag("zpig:in_invasion");
                    invader.addTag("zpig:invasion_id." + invasion.id);
                    invadersHealth += invader.getComponent("health").currentValue;
                });
                invasion.setProperty("zpig:num_invaders_in_wave", invaders.length);
                const health_boost_strength = Math.ceil(invadersHealth / 4);
                invasion.runCommand(`effect @s health_boost infinite ${health_boost_strength} true`);
                invasion.getComponent("health").resetToMaxValue();
            }
            else if (wave == 1 && wave_spawned == false && boss_bar_charged == 2 && !on_final_wave) { //spawn wave 2
                const pig1 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig2 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig3 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig4 = invasion.dimension.spawnEntity("zpig:piglin_runt<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig5 = invasion.dimension.spawnEntity("zpig:piglin_runt<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                invasion.dimension.playSound("portal.travel", pig1.location);
                invasion.setProperty("zpig:wave_spawned", true);
                const invaders = [pig1, pig2, pig3, pig4, pig5];
                let invadersHealth = 0;
                invaders.forEach((invader) => {
                    invader.addTag("zpig:in_invasion");
                    invader.addTag("zpig:invasion_id." + invasion.id);
                    invadersHealth += invader.getComponent("health").currentValue;
                });
                invasion.setProperty("zpig:num_invaders_in_wave", invaders.length);
                const health_boost_strength = Math.ceil(invadersHealth / 4);
                invasion.runCommand(`effect @s health_boost infinite ${health_boost_strength} true`);
                invasion.getComponent("health").resetToMaxValue();
            }
            else if (wave == 2 && wave_spawned == false && boss_bar_charged == 2 && !on_final_wave) { //spawn wave 3
                const pig1 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig2 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig3 = invasion.dimension.spawnEntity("zpig:piglin_runt<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig4 = invasion.dimension.spawnEntity("zpig:warboar<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                invasion.dimension.playSound("portal.travel", pig1.location);
                invasion.setProperty("zpig:wave_spawned", true);
                const invaders = [pig1, pig2, pig3, pig4];
                let invadersHealth = 0;
                invaders.forEach((invader) => {
                    invader.addTag("zpig:in_invasion");
                    invader.addTag("zpig:invasion_id." + invasion.id);
                    invadersHealth += invader.getComponent("health").currentValue;
                });
                invasion.setProperty("zpig:num_invaders_in_wave", invaders.length);
                const health_boost_strength = Math.ceil(invadersHealth / 4);
                invasion.runCommand(`effect @s health_boost infinite ${health_boost_strength} true`);
                invasion.getComponent("health").resetToMaxValue();
            }
            else if (wave == 3 && wave_spawned == false && boss_bar_charged == 2 && !on_final_wave) { //spawn wave 4
                const pig1 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig2 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig3 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig4 = invasion.dimension.spawnEntity("zpig:piglin_runt<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig5 = invasion.dimension.spawnEntity("zpig:spore_medic<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig6 = invasion.dimension.spawnEntity("zpig:spore_medic<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                invasion.dimension.playSound("portal.travel", pig1.location);
                invasion.setProperty("zpig:wave_spawned", true);
                const invaders = [pig1, pig2, pig3, pig4, pig5, pig6];
                let invadersHealth = 0;
                invaders.forEach((invader) => {
                    invader.addTag("zpig:in_invasion");
                    invader.addTag("zpig:invasion_id." + invasion.id);
                    invadersHealth += invader.getComponent("health").currentValue;
                });
                invasion.setProperty("zpig:num_invaders_in_wave", invaders.length);
                const health_boost_strength = Math.ceil(invadersHealth / 4);
                invasion.runCommand(`effect @s health_boost infinite ${health_boost_strength} true`);
                invasion.getComponent("health").resetToMaxValue();
            }
            else if (wave == 4 && wave_spawned == false && boss_bar_charged == 2 && !on_final_wave) { //spawn wave 5
                const pig1 = invasion.dimension.spawnEntity("zpig:piglin_runt<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig2 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig3 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig4 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig5 = invasion.dimension.spawnEntity("zpig:spore_medic<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig6 = invasion.dimension.spawnEntity("zpig:warboar<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig7 = invasion.dimension.spawnEntity("zpig:warboar<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                let pig8;
                if (hordeType == 0) {
                    pig8 = invasion.dimension.spawnEntity("zpig:seeker_piglin<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                }
                invasion.dimension.playSound("portal.travel", pig1.location);
                invasion.setProperty("zpig:wave_spawned", true);
                const invaders = [pig1, pig2, pig3, pig4, pig5, pig6, pig7, pig8];
                let invadersHealth = 0;
                invaders.forEach((invader) => {
                    invader.addTag("zpig:in_invasion");
                    invader.addTag("zpig:invasion_id." + invasion.id);
                    invadersHealth += invader.getComponent("health").currentValue;
                });
                invasion.setProperty("zpig:num_invaders_in_wave", invaders.length);
                const health_boost_strength = Math.ceil(invadersHealth / 4);
                invasion.runCommand(`effect @s health_boost infinite ${health_boost_strength} true`);
                invasion.getComponent("health").resetToMaxValue();
            }
            else if (wave == 5 && wave_spawned == false && boss_bar_charged == 2 && !on_final_wave) { //spawn wave 6
                const pig1 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig2 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig3 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig4 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig5 = invasion.dimension.spawnEntity("zpig:spore_medic<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig6 = invasion.dimension.spawnEntity("zpig:warboar<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                let pig7;
                let pig8;
                if (hordeType == 0) {
                    pig7 = invasion.dimension.spawnEntity("zpig:seeker_piglin<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    pig8 = invasion.dimension.spawnEntity("zpig:seeker_piglin<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                }
                invasion.dimension.playSound("portal.travel", pig1.location);
                invasion.setProperty("zpig:wave_spawned", true);
                const invaders = [pig1, pig2, pig3, pig4, pig5, pig6, pig7, pig8];
                let invadersHealth = 0;
                invaders.forEach((invader) => {
                    invader.addTag("zpig:in_invasion");
                    invader.addTag("zpig:invasion_id." + invasion.id);
                    invadersHealth += invader.getComponent("health").currentValue;
                });
                invasion.setProperty("zpig:num_invaders_in_wave", invaders.length);
                const health_boost_strength = Math.ceil(invadersHealth / 4);
                invasion.runCommand(`effect @s health_boost infinite ${health_boost_strength} true`);
                invasion.getComponent("health").resetToMaxValue();
            }
            else if (wave == 6 && wave_spawned == false && boss_bar_charged == 2 && !on_final_wave) { //spawn wave 7
                const pig1 = invasion.dimension.spawnEntity("zpig:piglin_runt<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig2 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig3 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig4 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig5 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig6 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig7 = invasion.dimension.spawnEntity("zpig:spore_medic<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig8 = invasion.dimension.spawnEntity("zpig:spore_medic<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig9 = invasion.dimension.spawnEntity("zpig:warboar<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                const pig10 = invasion.dimension.spawnEntity("zpig:warboar<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                let pig11;
                let pig12;
                let pig13;
                if (hordeType == 0) {
                    pig11 = invasion.dimension.spawnEntity("zpig:seeker_piglin<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    pig12 = invasion.dimension.spawnEntity("zpig:seeker_piglin<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    pig13 = invasion.dimension.spawnEntity("zpig:seeker_piglin<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                }
                invasion.dimension.playSound("portal.travel", pig1.location);
                invasion.setProperty("zpig:wave_spawned", true);
                const invaders = [pig1, pig2, pig3, pig4, pig5, pig6, pig7, pig8, pig9, pig10, pig11, pig12, pig13];
                let invadersHealth = 0;
                invaders.forEach((invader) => {
                    invader.addTag("zpig:in_invasion");
                    invader.addTag("zpig:invasion_id." + invasion.id);
                    invadersHealth += invader.getComponent("health").currentValue;
                });
                invasion.setProperty("zpig:num_invaders_in_wave", invaders.length);
                const health_boost_strength = Math.ceil(invadersHealth / 4);
                invasion.runCommand(`effect @s health_boost infinite ${health_boost_strength} true`);
                invasion.getComponent("health").resetToMaxValue();
            }
            else if (on_final_wave && wave_spawned == false && boss_bar_charged == 2) { //spawn final wave
                let bad_omen_level = invasion.getProperty("zpig:bad_omen_level");
                let invaders = [];
                let pig1;
                if (bad_omen_level <= 0) {
                    if (hordeType == 0) {
                        pig1 = invasion.dimension.spawnEntity("zpig:portal_guard<zpig:spawn_for_barbarian_invasion>", ValidLocationForWave(invasion));
                    }
                    const pig2 = invasion.dimension.spawnEntity("zpig:piglin_runt<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    const pig3 = invasion.dimension.spawnEntity("zpig:piglin_runt<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    invaders = [pig1, pig2, pig3];
                }
                else if (bad_omen_level == 1) {
                    if (hordeType == 0) {
                        pig1 = invasion.dimension.spawnEntity("zpig:portal_guard<zpig:spawn_for_barbarian_invasion>", ValidLocationForWave(invasion));
                    }
                    const pig2 = invasion.dimension.spawnEntity("zpig:piglin_runt<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    const pig3 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    const pig4 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    invaders = [pig1, pig2, pig3, pig4];
                }
                else if (bad_omen_level == 2) {
                    if (hordeType == 0) {
                        pig1 = invasion.dimension.spawnEntity("zpig:portal_guard<zpig:spawn_for_barbarian_invasion>", ValidLocationForWave(invasion));
                    }
                    const pig2 = invasion.dimension.spawnEntity("zpig:spore_medic<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    const pig3 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    const pig4 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    const pig5 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    invaders = [pig1, pig2, pig3, pig4, pig5];
                }
                else if (bad_omen_level == 3) {
                    if (hordeType == 0) {
                        pig1 = invasion.dimension.spawnEntity("zpig:portal_guard<zpig:spawn_for_barbarian_invasion>", ValidLocationForWave(invasion));
                    }
                    const pig2 = invasion.dimension.spawnEntity("zpig:spore_medic<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    const pig3 = invasion.dimension.spawnEntity("zpig:warboar<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    const pig4 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    const pig5 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    invaders = [pig1, pig2, pig3, pig4, pig5];
                }
                else if (bad_omen_level >= 4) {
                    if (hordeType == 0) {
                        pig1 = invasion.dimension.spawnEntity("zpig:portal_guard<zpig:spawn_for_barbarian_invasion>", ValidLocationForWave(invasion));
                    }
                    const pig2 = invasion.dimension.spawnEntity("zpig:portal_guard<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    const pig3 = invasion.dimension.spawnEntity("zpig:spore_medic<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    const pig4 = invasion.dimension.spawnEntity("zpig:spore_medic<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    const pig5 = invasion.dimension.spawnEntity("zpig:warboar<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    const pig6 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    const pig7 = invasion.dimension.spawnEntity("zpig:piglin_royal_soldier<zpig:spawn_for_invasion>", ValidLocationForWave(invasion));
                    invaders = [pig1, pig2, pig3, pig4, pig5, pig6, pig7];
                }
                invasion.dimension.playSound("portal.travel", pig1.location);
                invasion.setProperty("zpig:wave_spawned", true);
                let invadersHealth = 0;
                invaders.forEach((invader) => {
                    invader.addTag("zpig:in_invasion");
                    invader.addTag("zpig:invasion_id." + invasion.id);
                    invadersHealth += invader.getComponent("health").currentValue;
                });
                invasion.setProperty("zpig:num_invaders_in_wave", invaders.length);
                const health_boost_strength = Math.ceil(invadersHealth / 4);
                invasion.runCommand(`effect @s health_boost infinite ${health_boost_strength} true`);
                invasion.getComponent("health").resetToMaxValue();
            }
        }
        else if (!invasion.getProperty("zpig:invasion_over")) { //check for bad omen and trigger invasion
            const players = invasion.dimension.getPlayers({location: invasion.location, maxDistance: 25, excludeGameModes: ['spectator'],});
            players.forEach((player) => {
                if (player.getEffect("bad_omen") != undefined) {
                    invasion.setProperty("zpig:bad_omen_level", player.getEffect("bad_omen").amplifier);
                    invasion.runCommand("event entity @s zpig:reevaluate_difficulty");
                    invasion.runCommand("event entity @s zpig:start_invasion");
                    invasion.setProperty("zpig:invasion_horde_type", Math.floor(Math.random() * 0));
                    player.removeEffect("bad_omen");
                    const dimension = invasion.dimension;
                    const location = invasion.location;
                    if (dimension.getBlock(location).typeId == "zpig:finished_nether_reactor_core") {
                        dimension.getBlock(location).setPermutation(BlockPermutation.resolve("zpig:nether_reactor_core"));
                    }
                    dimension.getBlock(location).setPermutation(dimension.getBlock(location).permutation.withState("zpig:reactor_core_state", (1)));
                    if (dimension.getBlock({ x: location.x, y: location.y + 1, z: location.z }).typeId == "minecraft:obsidian") {
                        dimension.runCommand("setblock ~ ~2 ~ fire keep");
                    }
                }
            });
        }
    });
}

system.runInterval(InvasionLogic, 1);